#include "flexSerial.h"
#define BAUDRATE 115200
#define FLEXIO1_CLOCK (480000000L/16) // Again assuming default clocks?


bool FlexSerial::begin(uint32_t baud) {
	if (_txPin != -1) {
		_transmit_pflex = FlexIOHandler::mapIOPinToFlexIOHandler(_txPin, _transmit_flex_pin);
		if (_transmit_pflex == nullptr) {
			Serial.printf("FlexSerial - Failed to map TX pin %d to FlexIO\n", _txPin);
			return false;
		}

		// BUGBUG need to handle restarts...
		IMXRT_FLEXIO_t *p = &_transmit_pflex->port();
		Serial.printf("pin %d maps to: %x, port: %x pin %x\n", _txPin, (uint32_t)_transmit_pflex, (uint32_t)p, _transmit_flex_pin);
		
		_transmit_timer = _transmit_pflex->requestTimers();
		_transmit_shifter = _transmit_pflex->requestShifters();
		if ((_transmit_timer == 0xff) || (_transmit_shifter == 0xff)) {
			Serial.println("FlexSerial - Failed to allocate timer or shifter");
			return false;

		}

		_transmit_shifter_mask = 1;
		for (uint8_t i = _transmit_shifter; i > 0; i++) _transmit_shifter_mask <<= 1;

		Serial.printf("timer index: %d shifter index: %d mask: %x\n", _transmit_timer, _transmit_shifter, _transmit_shifter_mask);

		// lets try to configure a tranmitter like example
		Serial.println("Before configure flexio");
		p->SHIFTCFG[_transmit_shifter] = FLEXIO_SHIFTCFG_SSTOP(3) | FLEXIO_SHIFTCFG_SSTART(2); //0x0000_0032;
		p->SHIFTCTL[_transmit_shifter] = FLEXIO_SHIFTCTL_PINCFG(3) | FLEXIO_SHIFTCTL_SMOD(2) |
		                              FLEXIO_SHIFTCTL_TIMSEL(_transmit_timer) | FLEXIO_SHIFTCTL_PINSEL(_transmit_flex_pin); // 0x0003_0002;
		uint16_t baud_div =  (FLEXIO1_CLOCK/baud)/2 - 1;                                   
		p->TIMCMP[_transmit_timer] = 0xf00 | baud_div; //0xF01; //0x0000_0F01;		//
		p->TIMCFG[_transmit_timer] = FLEXIO_TIMCFG_TSTART | FLEXIO_TIMCFG_TSTOP(2) |
		                          FLEXIO_TIMCFG_TIMENA(2) |  FLEXIO_TIMCFG_TIMDIS(2); //0x0000_2222;
		p->TIMCTL[_transmit_timer] = FLEXIO_TIMCTL_TIMOD(1) | FLEXIO_TIMCTL_TRGPOL | FLEXIO_TIMCTL_TRGSRC
		                          | FLEXIO_TIMCTL_TRGSEL(1) | FLEXIO_TIMCTL_PINSEL(_transmit_flex_pin);  // 0x01C0_0001;
		p->CTRL = FLEXIO_CTRL_FLEXEN;
		//p->SHIFTSTAT = _transmit_shifter_mask;   // Clear out the status.

		// Set the IO pin into FLEXIO mode
		_transmit_pflex->setIOPinToFlexMode(_txPin);
		_transmit_pflex->addIOHandlerCallback(this);

		// Lets print out some of the settings and the like to get idea of state
		Serial.printf("CCM_CDCDR: %x\n", CCM_CDCDR);
		Serial.printf("VERID:%x PARAM:%x CTRL:%x PIN: %x\n", p->PARAM, p->CTRL, p->CTRL, p->PIN);
		Serial.printf("SHIFTSTAT:%x SHIFTERR=%x TIMSTAT=%x\n", p->SHIFTSTAT, p->SHIFTERR, p->TIMSTAT);
		Serial.printf("SHIFTSIEN:%x SHIFTEIEN=%x TIMIEN=%x\n", p->SHIFTSIEN, p->SHIFTEIEN, p->TIMIEN);
		Serial.printf("SHIFTSDEN:%x SHIFTSTATE=%x\n", p->SHIFTSDEN, p->SHIFTSTATE);
		Serial.printf("SHIFTCTL:%x %x %x %x\n", p->SHIFTCTL[0], p->SHIFTCTL[1], p->SHIFTCTL[2], p->SHIFTCTL[3]);
		Serial.printf("SHIFTCFG:%x %x %x %x\n", p->SHIFTCFG[0], p->SHIFTCFG[1], p->SHIFTCFG[2], p->SHIFTCFG[3]);
		Serial.printf("TIMCTL:%x %x %x %x\n", p->TIMCTL[0], p->TIMCTL[1], p->TIMCTL[2], p->TIMCTL[3]);
		Serial.printf("TIMCFG:%x %x %x %x\n", p->TIMCFG[0], p->TIMCFG[1], p->TIMCFG[2], p->TIMCFG[3]);
		Serial.printf("TIMCMP:%x %x %x %x\n", p->TIMCMP[0], p->TIMCMP[1], p->TIMCMP[2], p->TIMCMP[3]);
	}
	return true;
}

void FlexSerial::end(void) {
	// If the transmit was allocated free it now as well as timers and shifters.
	if (_transmit_pflex) {
		if (_transmit_timer != 0xff) {
			_transmit_pflex->freeTimers(_transmit_timer);
			_transmit_timer = 0xff;
		}
		if (_transmit_shifter != 0xff) {
			_transmit_pflex->freeShifters(_transmit_shifter);
			_transmit_shifter = 0xff;
		}

		_transmit_pflex->removeIOHandlerCallback(this);
		_transmit_pflex = nullptr;	
	}

}

int FlexSerial::available(void) {
	return 0;
}

int FlexSerial::peek(void) {
	return 0;
}

void FlexSerial::flush(void) {

}

size_t FlexSerial::write(uint8_t c) {
	uint32_t head;
	head = _txBuffer_head;
	if (++head >= TX_BUFFER_SIZE) head = 0;
	while (_txBuffer_tail == head) {
		/*
		int priority = nvic_execution_priority();
		if (priority <= hardware->irq_priority) {
			if ((port->STAT & LPUART_STAT_TDRE)) {
				uint32_t tail = _txBuffer_tail;
				if (++tail >= TX_BUFFER_SIZE) tail = 0;
				if (tail < tx_buffer_size_) {
					n = tx_buffer_[tail];
				} else {
					n = tx_buffer_storage_[tail-tx_buffer_size_];
				}
				port->DATA  = n;
				_txBuffer_tail = tail;
			}
		} else if (priority >= 256) 
		*/
		{
			yield(); // wait
		} 
	}
	//digitalWrite(5, LOW);
	//Serial.printf("WR %x %d %d %d %x %x\n", c, head, tx_buffer_size_,  TX_BUFFER_SIZE, (uint32_t)tx_buffer_, (uint32_t)tx_buffer_storage_);
	_txBuffer[_txBuffer_head] = c;
	__disable_irq();
	//transmitting_ = 1;
	_txBuffer_head = head;
	//port->CTRL |= LPUART_CTRL_TIE; // (may need to handle this issue)BITBAND_SET_BIT(LPUART0_CTRL, TIE_BIT);
	_transmit_pflex->port().SHIFTSIEN |= _transmit_shifter_mask;  // enable interrupt on this one...

	__enable_irq();
	//digitalWrite(3, LOW);
	return 1;
/*

	IMXRT_FLEXIO_t *p = &_transmit_pflex->port();

	while (!(p->SHIFTSTAT & _transmit_shifter_mask)) ; // wait until it says there is room for next output
	p->SHIFTBUF[_transmit_shifter] = c;  // put the next byte out
	return 1;
*/
}

int FlexSerial::read(void) {
	return -1;
}

void FlexSerial::clear(void) {
	//rx_buffer_head_ = rx_buffer_tail_;
}

int FlexSerial::availableForWrite(void) {
	uint32_t head, tail;

	head = _txBuffer_head;
	tail = _txBuffer_tail;
	if (head >= tail) return TX_BUFFER_SIZE - 1 - head + tail;
	return tail - head - 1;


}


bool FlexSerial::call_back (FlexIOHandler *pflex) {
	if (pflex == _transmit_pflex) {
		if (_transmit_pflex->port().SHIFTSTAT & _transmit_shifter_mask) {
			if (_txBuffer_head != _txBuffer_tail) {
				_transmit_pflex->port().SHIFTBUF[_transmit_shifter] = _txBuffer[_txBuffer_tail++] ;
				if (_txBuffer_tail >= TX_BUFFER_SIZE) _txBuffer_tail = 0;
			}
			if (_txBuffer_head == _txBuffer_tail) {
				_transmit_pflex->port().SHIFTSIEN &= ~_transmit_shifter_mask;  // disable interrupt on this one...
			}
		}
	}
	return false;  // right now always return false... 

}

