#include <Arduino.h>
#include "flex_io.h"
#ifndef _FLEX_SERIAL_H_
#define _FLEX_SERIAL_H_



class FlexSerial : public Stream, public FlexIOHandlerCallback
{
public:
	enum {TX_BUFFER_SIZE=64, RX_BUFFER_SIZE=40};
	FlexSerial(int rxPin=-1, int txPin=-1, bool inverse_logic = false) :
		_rxPin(rxPin), _txPin(txPin) {};

	~FlexSerial() { end(); }
	bool begin(uint32_t baud);
	void end(void);

	virtual int available(void);
	virtual int peek(void);
	virtual void flush(void);
	virtual size_t write(uint8_t c);
	virtual int read(void);
	using Print::write;

	void clear(void);
	int availableForWrite(void);

	// Call back from flexIO when ISR hapens
	virtual bool call_back (FlexIOHandler *pflex);


private:
	int _rxPin;
	int _txPin;

	// Variables for tranmitter
	FlexIOHandler *_transmit_pflex = nullptr;
	uint8_t 		_transmit_flex_pin;
	uint8_t 		_transmit_timer = 0xff;
	uint8_t 		_transmit_shifter = 0xff;
	uint8_t 		_transmit_shifter_mask;
	uint8_t			_txBuffer[TX_BUFFER_SIZE];
	uint16_t		_txBuffer_head = 0;
	uint16_t 		_txBuffer_tail = 0;

};
#endif //_FLEX_SERIAL_H_